########################################################################
## Wed May 12 10:16:42 PDT 2010 Finding the vital demographic rates
## for a population that never was.  In this exercise we ask, and
## attempt to answer the question: What rates are consistent with a
## certain observed fact about a population.  Specifically, suppowe we
## know only that the mean number of living maternal ancestors
## (mother, grandmonter, great grandother ...) of a person at birth
## was 2.  What fertility and mortality rates might have given rise to
## this condition, which we refer to hereafter as "Josh's G = 2.0".
## The approach we take will be to run Socsim systematically over a
## grid of possible rates and for each set of rates, to calculate the 
## likelihood that those rates would produce a value of Josh's G = 2.0.

########################################################################
## Section 1:  some functions that will be convenient to call later
########################################################################

## setting scipen will tell R to avoid using scientific notation
options(scipen=999)

## If you are running under windows with cygwin & etc...
## then set windows<-TRUE. This affects writeScaledFert,writeScaledMort, and
## Likelihood_simul.  the application "fromdos.exe"  is required.

windows<-FALSE

## acquire a matrix of Coale-Demeny nqx over which to interpolate

library(demogR) ## Hope that's installed, if not
## install.packages("demogR") should work
## if that's not practical try
## load("CDmort.rsave")
## and skip the next few lines ... all the
## way to the next function definition.

## create a list whose elements are:
## (1) a matrix of nqx values columns are age
## categories and whose rows refer to different levels of coale-demeny model
## life table West
## (2) a vector of upper age bounds of the age categories in the nqx
## matrix (the column) these are what Socsim expects in rate files
## (3) a vector of e0 coresponding to the rows of the the nqx matrix.

## The point of this list is to allow us to crudely interpolate nqx
## vectors for any e0.

CDmort<-  list(nqx={nqx<-cdmltw(sex="F")$nqx
                    nqx},
               mo=c(as.numeric(dimnames(nqx)[[2]])[-1],100)*12,
               e0=cdmltw(sex="F")$ex[,1])

rownames(CDmort$nqx)<-paste(CDmort$e0)

##save(file="CDmort.rsave",CDmort)

##load(file="CDmort.rsave")

########################################################################
writeScaledFert<-function(outfile="../fert_rates_rescaled",
                          scaleFactor=26){
  
  ## called by likelihood_simul before running a bunch of
  ## socsimulations this will write a legal file of fertility rates
  ## scaleFactor should be about the mean age at childbirth ignoring
  ## mortality we'll do this by letting fertility be dnorm with
  ## scaleFactor being the mean.

  ## The gist is: this function writes fertility rate files that are
  ## scaled to achieve a given mean age at child bearing.  The rates
  ## have the shape of a normal density function.

  bprob<-dnorm(x=15:50,mean=scaleFactor,sd=2)
  Srate<-cbind(15:50,0,bprob*(2.1/12))
  Srate<-rbind(Srate,c(100,0,0))
  cat(file=outfile,
      x="*Rescaled fertility rates written by writeScaledFert in  calibrate.r\n",
      append=F)
  ##male rates
  cat(file=outfile,x=paste('*',"rescale parameter ~ mean age child bearing",
                     scaleFactor),append=T)
  cat(file=outfile,x="\nbirth 1 F married 0\n",append=T)

  write.table(file=outfile,x=Srate,append=T,
              quote=FALSE,row.names=FALSE,col.names=FALSE) 

  ## divorced and widowed will default to single rate so only married
  ## and single need to be specified
  
  singleRates<-data.frame(yr=100,mo=0,rate=0)

  cat(file=outfile,x="\nbirth 1 F single 0\n",append=T)
  write.table(file=outfile,x=Srate,append=T,
              quote=FALSE,row.names=FALSE,col.names=FALSE) 
  ## must convert to unix format when in windows
  if(windows){
    shell(paste("fromdos.exe", outfile,sep=" "),"cmd")
  }
}

########################################################################
writeScaledMort<-function( outfile="../mort_rates_rescaled",
                          Mort=CDmort,
                          scaleFactor=69){
  ## interpolates C-D model life table mort rates for e0= scale factor
  ## then writes them in socsim ready format to outfile.  More
  ## complicated than writeScaledFert but the same idea: this writes a
  ## legal socsim mortality rate file where the .sup file says it will
  ## be.
  ## the mortality rates are Coale-Demeny rates or interpolations
  ## corresponding to the given e0 (scaleFactor).

  ## scaleFactor must be in e0 range
  stopifnot(max(Mort$e0) >= scaleFactor, min(Mort$e0)<=scaleFactor)


  ## rownames of nqx should == e0 but sometimes rounding error undoes
  ## the beauty of science
  ## lower<- max(Mort$e0[Mort$e0 <= scaleFactor])
  ##  upper<- min(Mort$e0[Mort$e0 >= scaleFactor])
  e0<-as.numeric(rownames(Mort$nqx))
  lower<- max(e0[e0 <= scaleFactor])
  upper<- min(e0[e0 >= scaleFactor])
  
  loc<-(scaleFactor-lower)/(upper-lower)
  
  nqx<-Mort$nqx[paste(lower),]*loc +  Mort$nqx[paste(upper),]*(1-loc)
  ## convert to monthly prob
  width<-diff(c(0,Mort$mo))
  prob<- 1-(1-nqx)^(1/width)
  Srates<-cbind(0,Mort$mo,prob)

  if(max(Mort$mo) < 1200){
    ## make sure eveyone dies at age 100 to avoid warning message from Socsim
    Srates<-rbind(Srates,c(0,1200,1))
  }

  cat(file=outfile,
      x="*Rescaled mortality rates written by writeScaledMort() in  calibrateG.r\n",
      append=F)
  
  
  ##male rates
  cat(file=outfile,x=paste('*',"rescale parameter (~e0)", scaleFactor),append=T)
  cat(file=outfile,x="\ndeath 1 M single\n",append=T)
  write.table(file=outfile,x=Srates,append=T,
              quote=FALSE,row.names=FALSE,col.names=FALSE) 
  
  ##female rates (no difference between male and female for now)
  cat(file=outfile,x="\ndeath 1 F single\n",append=T)
  write.table(file=outfile,x=Srates,append=T,
              quote=FALSE,row.names=FALSE,col.names=FALSE) 
  ## must convert to unix format when in windows
  if(windows){
    shell(paste("fromdos.exe", outfile,sep=" "),"cmd")
  }
}  
########################################################################

getG <-function(opopfile="../SimResults/pop_output.opop"){

  ## called by liklighood_simul, this will read the opop file
  ## expecting to find a socsim opop, and will calculate G = number of
  ## living maternal ancestors for the people born in the last month
  ## of the simulation

  momAlive<-function(pid,sofar=0){
    ## assumes that opop is in the environment which it is if this
    ## function is left inside of getG
    
    if(is.na(opop$mom[pid])|opop$mom[pid]==0){
      return(sofar)
    }else{
      sofar<-sofar+(opop$dod[opop$mom[pid]]==0)
      ## the function calls itself cool huh?
      momAlive(opop$mom[pid],sofar)
      
    }
  }

  opop <- read.table(file=opopfile,header=F,as.is=T)
  ## Rename the columns of the opop file
  names(opop)<-c("pid","fem", "group", "nev", "dob", "mom",
                 "pop", "nesibm", "nesibp", "lborn",
                 "marid", "mstat", "dod",
                 "fmult" )
  
  ##current period
  endmo <- max(opop$dob)

  lastCohort<-opop$pid[opop$dob>=(endmo-59)] ## include all born in last 5 yrs
  
  return(mean(sapply(lastCohort,momAlive)))
}
  

##

##
######################################################################## 
likelihood_simul <- function(k,num_runs=10, targetQofI=2.0,
                             fertFile="fert_rates_rescaled",
                             mortFile="mort_rates_rescaled",
                             Mort=CDmort ,
                             Fert=birthUnscaled,
                             tmpdir="00"){

  ## This function create and populates a temporary directory in order
  ## that it may run in paralell via %dopar% If /72hours does not exist
  ## in your filesystem  perhaps use /tmp instead but beware of filling /tmp
  User<-system("whoami",intern=TRUE)
  TmpDir<-paste("/72hours",User,tmpdir,sep="/")
  system(paste("mkdir -p ",TmpDir))
  system(paste("/bin/cp  -R ../", TmpDir,sep=" "))
######
  ##  This function runs socsim num_runs times, using rates determined
  ## by the vecor k of scalefactors. After each socsim run, it calls
  ## getG() which  reads
  ## the resulting output files; calculates and stores the Quantity of
  ## Interest, in this case Josh's G. It returns a list which includes
  ## the likelihood of Socsim producing a Quantity of Interest =
  ## targetQofI when using the rates implied by the values of k.
  
  ##  takes as input:

  ##  k: a vector of 2 elements, which contains rescaling factors for
  ## age-specific fertility and mortality rates, respectively
  
  ## num_runs: the number of runs, with different seeds, that we want
  ## to have for each combination of parameters, to estimate the
  ## likelihood

  ## targetQofI: the true known quantity of interest on which the
  ## likelihood calculations are based.

  
  
  names(k)<-c("FertScale","MortScale")

  ## Step 1 Rescale the fertility and mortality rates by the k parameters
  ## write to files where socsim expects to find rates
  
  writeScaledFert(outfile=paste(TmpDir,fertFile,sep='/'),
                  scaleFactor=k["FertScale"])
  
  
  writeScaledMort(outfile=paste(TmpDir,mortFile,sep='/'),
                  Mort=Mort,
                  scaleFactor=k["MortScale"])
  
  
  ##Step 2: run the simulation num_runs times, after each run,
  ##calculate and store the quantity of interest, for now this is
  ##Josh's G

  
  QofI<-NULL # will hold vector of quantities of interest, for now G
                                        
  
  for (index in 1:num_runs){
    
    ## RUN SOCSIM
    ## Windows version for Rostock MPI 116 June 2010
    if(windows){
      terminal_command<- paste("cd ..&&","N:\\socsim\\bin\\socsim",
                               " calibration_exercise.sup ",
                               round(runif(1,min=1,max=10000)), sep=" ")
      shell(terminal_command,"cmd")
    }else{
     ## Unix version
      terminal_command<- paste("cd ",TmpDir,";","socsim",
                             " calibration_exercise.sup ",
                               sample(1:999999,size=1),sep="")
##                              round(runif(1,min=1,max=10000)), sep="")
      system(terminal_command)
    }
    
    ##ANALYZE THE OUTPUT
    QofI <-c(QofI,
             ## This is where we find G. some other function that
             ## read socsim output and returned some other number
             ## could be substituted.
             getG(opopfile=
                  paste(TmpDir,"/SimResults/pop_output.opop",sep=''))
             )
    
  }
  
  ## collect valuable information on the quantity of interest
  meanQofI <- mean(QofI,na.rm=T)
  ## if there is no variation in QofI we have a train wreck.
  sd_QofI <- ifelse(sd(QofI,na.rm=T) !=0,
                    sd(QofI,na.rm=T),
                    0.000001)

  
  ##standardization of G
 # standard_targetQofI <- (targetQofI - meanQofI)/sd_QofI
  ## compute the likelihood based on a normal approximation and
  ## deference to the Central Limit Theorem

##  like <- (exp(-((targetQofI - meanQofI)^2)/(2*(sd_QofI)^2))*
##           ((2*pi*(sd_QofI)^2)^(-0.5))) 

  like<- dnorm(x=targetQofI,mean=meanQofI,sd=sd_QofI)
  result <- list(likelihood=like,
                 targetQofI=targetQofI,
                 k=k,
                 meanQofI=meanQofI,
                 sdQofI=sd_QofI,
                 simulatedQofI= QofI)
  return(result)
  
}

########################################################################

########################################################################
## Using the functions defined above... let's see what we can do
########################################################################

## Let's run likelihood_simul to do 10 simulations

like10<-likelihood_simul(k=c(26,80),
                        num_runs=10,
                         targetQofI=2.0,
                         tmpdir="test1")


like10

## NOTE that there are now 10 log files cluttering Calibration directory
## these files are called calibration_exercise.supNNN.log. At a convenient time##  such as when they get in your way, delte them
## shell("del ../*.log","cmd")  ## in windows or
## system("\rm ../*log") ## in Unix  

## histogram of the simulated Gs. this is the "quantity of
## interest". We know that in Imaginatia this is 2.0 From 10 trials
## of Socsim we get 10 different estimated Gs:

stem(like10$simulatedQofI)

## Let's look at the estimated density to get a nicer picture

##postscript("density_rescaled.eps",horizontal=FALSE)
hist(like10$simulatedQofI,
     ylim=c(0,max(density(like10$simulatedQofI,adjust=2)$y)),
     xlim=c(1.5,3),
     xlab="Josh's G in Simulated Populations",
     main=paste("Rescaling Parameters: Fertility =",like10$k["FertScale"],
       "Mortality =", like10$k["MortScale"]) )

lines(density(like10$simulatedQofI,adjust=2)$x,
      density(like10$simulatedQofI,adjust=2)$y,
      type="l",xlab="",ylab="",
      col="red")

abline(v=like10$targetQofI,lty=2,lwd=2)


## let's see how much things improves if we run 100 simulations
## instead of just 10.  Uncomment and run the line below to find
## out. It will take a few minutes so maybe get coffee OR IF YOU'RE in
## a hurry then you can load a pre-cooked like100.  The like100.rsave
## file was run in a Linux OS so it should work fine under
## windows.

### Don't uncomment unless you want to run Socsim 100 times and save the output
##like100 <- likelihood_simul(k=c(26,80), num_runs=100,targetQofI= 2.0) 
##save(file="like100.rsave",like100)

## run this to save time  
load(file="like100.rsave")

lines(density(like100$simulatedQofI,adjust=2)$x,
      density(like100$simulatedQofI,adjust=2)$y,
      type="l",
      col="blue")


#dev.off()

########################################################################
## So now we can evaluate the likelihood that a particular set of
## mortality and feritlity rates will produce the target quantity of
## interest.  Our goal, however, is to find the set of rates that is
## most likely to produce simulated populations with the known target
## quanitity (Josh's G = 2.0) To do that, we need to evaluate the
## likelihood for a bunch of rate sets and then interpolate over the
## likelihood surface to find the maximum.
########################################################################

## Calculate the likelihood (that a set of rates will produce the
## target G) for the rates corresponding to a 10X10 grid of
## scale factors.

paramLength <- 10  ## dimension of the grid
kFert <- seq(22,29,length=paramLength) 
kMort <- seq(50,80,length=paramLength)

##create a matrix whose rows represent all the possible combinations
##of parameters
pointsGrid<- expand.grid(kFert,kMort)
names(pointsGrid)<- c("kFert","kMort")
pointsGrid$likely<-NA  ## place holder
pointsGrid$meanQofI<-NA
pointsGrid$sdQofI<-NA

##fill in the values of the likelihood
## Because we don't have all day, we'll do 5 runs for each pair of
## You can reduce the num_runs parameter further if you're the impatient sort.
## this does take a while. Sorry.

## BEFORE YOU PROCEDE note that you can do this later...
## uncomment the load() statement below to read a precooked version
## of pointsGrid -- it will save you 20 minutes.


## execute this for() loop ONLY if you plan to get coffee...
## parallel processing will speed this up ... a little

NAPTIME<-FALSE
if(NAPTIME){
  ## let's not run this by accident it will take hours to run!!!
  library(foreach)
  library(doMC)

  registerDoMC(cores=5) # register the multicore backend
  getDoParWorkers() #see how many cores can be used symultaneously
  
  PGresults<-foreach(line = 1:nrow(pointsGrid)) %dopar% {
    tdir<-paste('junk',line,sep='')
    res<-
      likelihood_simul(k=c(pointsGrid$kFert[line],
                         pointsGrid$kMort[line]),
                       num_runs=5,
                       targetQofI=2.0,
                       tmpdir=tdir
                       )
    ##  pointsGrid[line,3:5]<-c(res$likelihood,res$meanQofI,res$sdQofI)
    return(c(res$likelihood,res$meanQofI,res$sdQofI))
  }
  ## glue the results of all those socsim runs onto pointGrid so that
  ## the scale factors that were input will be lined up with the
  ## results
  lksim.res<-t(sapply(PGresults,c))
  pointsGrid[,3:5]<-lksim.res
  save(file="pointsGrid.rsave",pointsGrid)
}





## read a precooked version of pointsGrid
##
load(file="pointsGrid.rsave")


## Let's draw an exceptionally cool graph of the likelihood surface

pGSorted <- pointsGrid[order(pointsGrid$kMort,pointsGrid$kFert),]

### a matrix suitable for use in filled.contour, this one is for
### the G surface
meanMat <- matrix(pGSorted$meanQofI,
                   nrow=Npoints<-length(unique(pGSorted$kMort)),
                   ncol=Npoints,
                   byrow=T)


## this one is for the likelihood
likeMat <- matrix(pGSorted$likely,
                   nrow=Npoints<-length(unique(pGSorted$kMort)),
                   ncol=Npoints,
                   byrow=T)

#### find the points if any in pointsGrid where the mean across all
#### those simulations, of G =2.0.
JG2<-pointsGrid[pointsGrid$meanQofI == 2,]
## or maybe closest to 2
if(nrow(JG2)==0){
  JG2<-pointsGrid[abs(2 - pointsGrid$meanQofI) ==
                  min(abs(2 - pointsGrid$meanQofI)),]
}
  ##pdf(file="JoshGcontours.pdf")

### plot the surface of mean simulated Josh's G over the various mortality
### and fertility rates
filled.contour(unique(pGSorted$kMort),
               unique(pGSorted$kFert),
               meanMat,
               ylab="Rescaling parameter for fertility rates (~T)",
               xlab="Rescaling parameter for mortality rates (~e0)",
               main="Mean of Josh's G",
               levels=c(1,1.25,1.5,1.75,1.9,1.95,2.05,2.1,2.25,2.5,2.75,3,3.5),
               plot.axes={axis(1);axis(2);
                          points(x=JG2$kMort,
                                 y=JG2$kFert)})


### Now the same thing for the *likelihood* of producing a Josh's G=2.0

## find the scale factors that corespond to the fertility and
## mortality rates that have the highest **likelihood** of producing Socsim
## population with a G equal to the known true G in
## Imaginatia.  

## select the maximum likelihood points
MaxLkhood<- pointsGrid[pointsGrid$likely == max(pointsGrid$likely,na.rm=T),]
MaxLkhood
JG2
## compare the MaxLkhood point(s) to the the JG2 points.  They could
## be the same, since the likelihood calculation is based on the
## target value of G=2.0, BUT they could also be different since the
## sd of the simulated G values also plays in important part.

## MaxLkhood is really what we are after. If this were a real research
## project, we could publish those rates along with the graph below:

filled.contour(unique(pGSorted$kMort),
               unique(pGSorted$kFert),
               likeMat,
               ylab="Rescaling parameter for fertility rates (~T)",
               xlab="Rescaling parameter for mortality rates (~e0)",
               main="Likelihood contour",
               nlevels=15,
               plot.axes={axis(1);axis(2);
                          points(x=MaxLkhood$kMort,
                                 y=MaxLkhood$kFert)})

##maximum likelihood estimates of the parameters


##dev.off()

## in case we have to publish in a journal that is too cheap to print color...
contour(unique(pGSorted$kMort),
        unique(pGSorted$kFert),
        likeMat,
        ylab="Rescaling parameter for fertility rates (~T)",
        xlab="Rescaling parameter for mortality rates (~e0)",
        main="Likelihood contour",
        )

points(x=MaxLkhood$kMort, y=MaxLkhood$kFert,pch=4,cex=2,col='blue')
##dev.off()
########################################################################
## Extensions
########################################################################

########################################################################
## 1) Above we investigated G for rates that are at least reasonable
## for human populations.  Some scientists believe that Josh's G is
## pretty close to 2.0 for a surprisingly wide variety of other
## organisms.  Repurpose the above code to test a more arbitrary set
## of rates consider combinations of scensecent and nonscenesent
## mortality and fertility.

## scenesent mortality would be something like l(x) ~ gompertz -- or
## the same model life table rates that we used above.  nonscenest,
## could be l(x) ~ exponential (constant prob of mortality such as
## predation)

## scenescent fertility would be a declining function in age. Any
## 1-CDF could work as could something simpler.  nonscenesent
## fertility would be constant.
##
## Suggestion -- as long as marriage is not part of the story, you
## might considier treating Socsim months as years. Thow would shorten
## the time required to run the simulations and thereby allow you to
## run more of them. This is not worth the effort unless your simulations
## are longer than the ones in the example
########################################################################

########################################################################
## 2) (this is a long one)Suppose that, one of the referee's who read
## our paper harbors the foolish prejudice (excuse me, "prior") that
## the true mortality and fertilty rates in Imaginatia were such that
## e0 = 2*the mean age at child bearing and that her certainty about
## this declined in probability with with the exponential (rate=1) of
## |e0 -2* MeanAgeCB|.

##We should probably just submit the paper to a better journal but
## just to show off I mean to address the referee's concerns, we can
## instead calculate  the "posterior" distribution of the scale
## factors and then (maybe) rerun our simulation using the mean of the
## resulting distributions instead of the maximum likelihood estimates
## that we just produced.

## This is a pretty straightforward (though perhaps pointless)
## exercise of Bayes' Theorem.  We have evidence in the form of the
## likelihood surface, but we also have "prior" beliefs from the
## venerable referee.  These can be cleverly combined using the
## "SIR" algorithm.


## Step 0) Plot the "prior" distribution so we can see what the
## referee is talking about

##  referee says scale factors are such that e0=2*MeanAgeCB declining 
## in probability with increasing absolute value of e0-2*MeanAgeCB


pGSorted$priorProb<-with(pGSorted,
                         dexp(x=abs(kMort - 2*kFert), rate=1))


## here's what the prior looks like:
priorMat <- matrix(pGSorted$priorProb,
                   nrow=Npoints<-length(unique(pGSorted$kMort)),
                   ncol=Npoints,
                   byrow=T)


filled.contour(unique(pGSorted$kMort),
               unique(pGSorted$kFert),
               priorMat,
               ylab="Rescaling parameter for fertility rates",
               xlab="Rescaling parameter for mortality rates",
               main="Contour of prior on scaling parameters",
               plot.axes={axis(1);axis(2);
                          points(x=MaxLkhood$kMort,
                                 y=MaxLkhood$kFert)})


## Step 1) Sample with replacement a number m of parameter vectors (for
## which we have Socsim simulation data) from the "prior" distribution
## (that is according to the prejudices of the referee)

priorSample<- pGSorted[sample(1:nrow(pGSorted),1000,replace=TRUE,
                              prob=pGSorted$priorProb),1:3]


## Step 2) 
##Compute the sampling importance weights based on the likelihood 
priorSample$weights<-priorSample$likely/sum(pGSorted$likely)


## Step 3) Sample with replacement from the m-sized sample of
## parameter vectors (which were drawn according to the prior
## distribution) using the sampling importance weights as
## probabilities

posteriorParams<-priorSample[sample(1:nrow(priorSample),10000,replace=TRUE,
                                    prob=priorSample$weights),]

par(mfrow=c(1,2))
hist(posteriorParams$kFert,
     breaks=10,
     xlab="Rescaling parameter for fertility rates",
     main="Posterior: Fertility parameter")

hist(posteriorParams$kMort,
     breaks=10,
     xlab="Rescaling parameter for mortality rates",
     main="Posterior: Mortality parameter")




## Step 4) take the means of the posterior distribution and see what
## socsim tells us about.  
maxPostFert<- mean(posteriorParams$kFert)
maxPostMort<- mean(posteriorParams$kMort)

## compare this point -- the mean of the posterior to the maximum
## likelihood point(s) that we previously discovered.

c(maxPostFert,maxPostMort)
MaxLkhood

## Let's run it 25 times (or some other number of times if you are impatient)
## and see  if we get the G that we like and with what sort of likelihood
maxPostLkhood <-
  likelihood_simul(k=c(maxPostFert,maxPostMort),
                   num_runs=25, 
                   targetQofI=2.0,
                   tmpdir='ref') 


## and the likelihhod of producing G=2.0
par(mfrow=c(1,1))
hist(maxPostLkhood$simulatedQofI,
               xlab="Simulated population G",
               main=paste("Rescaling Parameters: Fertility =",
            round(maxPostFert,3)," Mortality =",round(maxPostMort,3)),
     sub="scaling parameters from SIR with prior")
     

with(maxPostLkhood,
     lines(density(simulatedQofI),
           type="l",
           col="red",
           ylab="Density")
          )

abline(v=2.0,lty=2,lwd=2)
## prejudice validated by science.

## Now getting to the point, as an extension you might try specifying
## a different prior.

########################################################################
## 3) One could calibrate socsim rates based on any population
##     measure.  population growth rate or median age for example. To
##     do this you simply write a different function and put in the
##     place of getG() in the likelihood_simul() function.
##
